
/* ------------------------ iem_anything ---------------------------- */
static t_class *iem_anything_class;

typedef struct _iem_anything
{
    t_object   x_obj;
    int        x_n;
    int        x_ac;
    t_atom     *x_at;
    t_symbol   *x_sym;
    t_atomtype x_type;
} t_iem_anything;

static void iem_anything_bang(t_iem_anything *x)
{
    if(x->x_type == A_COMMA)
        outlet_anything(x->x_obj.ob_outlet, x->x_sym, x->x_ac, x->x_at);
    else if(x->x_type == A_GIMME)
        outlet_list(x->x_obj.ob_outlet, &s_list, x->x_ac, x->x_at);
    else if(x->x_type == A_FLOAT)
        outlet_float(x->x_obj.ob_outlet, atom_getfloat(x->x_at));
    else if(x->x_type == A_SYMBOL)
        outlet_symbol(x->x_obj.ob_outlet, atom_getsymbol(x->x_at));
    else if(x->x_type == A_NULL)
        outlet_bang(x->x_obj.ob_outlet);
    else if(x->x_type == A_POINTER)
        outlet_pointer(x->x_obj.ob_outlet, (t_gpointer *)x->x_at->a_w.w_gpointer);
}

static void iem_anything_float(t_iem_anything *x, t_float f)
{
    x->x_ac = 1;
    SETFLOAT(x->x_at, f);
    x->x_sym = &s_float;
    x->x_type = A_FLOAT;
}

static void iem_anything_symbol(t_iem_anything *x, t_symbol *s)
{
    x->x_ac = 1;
    SETSYMBOL(x->x_at, s);
    x->x_sym = &s_symbol;
    x->x_type = A_SYMBOL;
}

static void iem_anything_pointer(t_iem_anything *x, t_gpointer *gp)
{
    x->x_ac = 1;
    SETPOINTER(x->x_at, gp);
    x->x_sym = &s_pointer;
    x->x_type = A_POINTER;
}

static void iem_anything_list(t_iem_anything *x, t_symbol *s, int ac, t_atom *av)
{
    t_atom *at;

    if(ac > x->x_n)
    {
        if(x->x_at)
            freebytes(x->x_at, x->x_n * sizeof(t_atom));
        x->x_n = ac;
	x->x_at = (t_atom *)getbytes(x->x_n * sizeof(t_atom));
    }
    x->x_ac = ac;
    x->x_sym = &s_list;
    at = x->x_at;
    while(ac--)
	*at++ = *av++;
    x->x_type = A_GIMME;
}

static void iem_anything_anything(t_iem_anything *x, t_symbol *s, int ac, t_atom *av)
{
    t_atom *at;

    if(ac > x->x_n)
    {
        if(x->x_at)
            freebytes(x->x_at, x->x_n * sizeof(t_atom));
        x->x_n = ac;
	x->x_at = (t_atom *)getbytes(x->x_n * sizeof(t_atom));
    }
    x->x_ac = ac;
    x->x_sym = s;
    at = x->x_at;
    while(ac--)
	*at++ = *av++;
    x->x_type = A_COMMA;
}

static void iem_anything_free(t_iem_anything *x)
{
    if(x->x_at)
        freebytes(x->x_at, x->x_n * sizeof(t_atom));
}

static void *iem_anything_new(t_symbol *s, int ac, t_atom *av)
{
    t_iem_anything *x = (t_iem_anything *)pd_new(iem_anything_class);
    int i;

    x->x_type = A_NULL;
    if(!ac)
    {
        x->x_type = A_NULL;
	x->x_sym = &s_bang;
        x->x_n = 1;
        x->x_ac = 0;
	x->x_at = (t_atom *)getbytes(x->x_n * sizeof(t_atom));
    }
    else if(ac == 1)
    {
	if(IS_A_SYMBOL(av,0))
	{
            x->x_type = A_COMMA;
	    x->x_sym = atom_getsymbol(av);
            x->x_n = 1;
            x->x_ac = 0;
	    x->x_at = (t_atom *)getbytes(x->x_n * sizeof(t_atom));
	}
	else
	{
	    if(IS_A_FLOAT(av,0))
            {
                x->x_type = A_FLOAT;
		x->x_sym = &s_float;
	    }
            else if(IS_A_POINTER(av,0))
            {
                x->x_type = A_POINTER;
		x->x_sym = &s_pointer;
	    }
            x->x_n = x->x_ac = 1;
	    x->x_at = (t_atom *)getbytes(x->x_n * sizeof(t_atom));
	    x->x_at[0] = *av;
	}
    }
    else
    {
	if(IS_A_SYMBOL(av,0))
	{
            x->x_type = A_COMMA;/*for anything*/
	    x->x_sym = atom_getsymbol(av++);
	    ac--;
	}
	else
        {
            x->x_type = A_GIMME;
	    x->x_sym = &s_list;
	}
        x->x_n = x->x_ac = ac;
        x->x_at = (t_atom *)getbytes(x->x_n * sizeof(t_atom));
        for(i=0; i<ac; i++)
	    x->x_at[i] = *av++;
    }
    outlet_new(&x->x_obj, &s_list);
    return (x);
}

static void iem_anything_setup(void)
{
    iem_anything_class = class_new(gensym("iem_anything"), (t_newmethod)iem_anything_new,
			       (t_method)iem_anything_free, sizeof(t_iem_anything), 0, A_GIMME, 0);
    class_addcreator((t_newmethod)iem_anything_new, gensym("ia"), A_GIMME, 0);
    class_addcreator((t_newmethod)iem_anything_new, gensym("any"), A_GIMME, 0);
    class_addbang(iem_anything_class, (t_method)iem_anything_bang);
    class_addanything(iem_anything_class, iem_anything_anything);
    class_addlist(iem_anything_class, iem_anything_list);
    class_addpointer(iem_anything_class, iem_anything_pointer);
    class_addfloat(iem_anything_class, (t_method)iem_anything_float);
    class_addsymbol(iem_anything_class, iem_anything_symbol);
    class_sethelpsymbol(iem_anything_class, gensym("iemhelp/help-iem_anything"));
}

